/// <reference types="node" />
/// <reference types="node" />
/// <reference types="node" />
import { Socket } from 'net';
import { DatabaseError, DatabaseTransaction, SelectorState } from '@deepkit/orm';
import { ReceiveType } from '@deepkit/type';
import { TLSSocket } from 'tls';
import { Host, PostgresClientConfig } from './config.js';
import { DefaultPlatform, PreparedEntity } from '@deepkit/sql';
export declare class PostgresError extends DatabaseError {
}
export declare class PostgresConnectionError extends PostgresError {
}
export interface Result {
    rows: any[];
    rowCount: number;
}
export declare class PostgresClientPrepared {
    client: PostgresClientConnection;
    sql: string;
    selector: SelectorState;
    private statementName;
    created: boolean;
    cache?: Buffer;
    deserialize: (message: Uint8Array, rows: any[]) => void;
    constructor(client: PostgresClientConnection, sql: string, selector: SelectorState, statementName: string);
    execute(params: any[]): Promise<Result>;
}
export declare function readMessageSize(buffer: Uint8Array | ArrayBuffer, offset?: number): number;
export declare class ResponseParser {
    protected readonly onMessage: (response: Uint8Array) => void;
    protected currentMessage?: Uint8Array;
    protected currentMessageSize: number;
    constructor(onMessage: (response: Uint8Array) => void);
    feed(data: Uint8Array, bytes?: number): void;
}
export declare enum ConnectionStatus {
    pending = "pending",
    connecting = "connecting",
    connected = "connected",
    disconnected = "disconnected"
}
export interface ConnectionRequest {
    readonly: boolean;
    nearest: boolean;
}
export declare const enum HostType {
    primary = 0,
    secondary = 1
}
export type TransactionTypes = 'REPEATABLE READ' | 'READ COMMITTED' | 'SERIALIZABLE';
export declare class PostgresDatabaseTransaction extends DatabaseTransaction {
    connection?: PostgresClientConnection;
    setTransaction?: TransactionTypes;
    /**
     * This is the default for mysql databases.
     */
    repeatableRead(): this;
    readCommitted(): this;
    serializable(): this;
    begin(): Promise<void>;
    commit(): Promise<void>;
    rollback(): Promise<void>;
}
type Deserializer = (message: Uint8Array) => any;
export declare abstract class Command<T> {
    current?: {
        resolve: Function;
        reject: Function;
        deserializer: Deserializer;
    };
    write: (buffer: Uint8Array) => void;
    sendAndWait(message: Uint8Array, deserializer?: Deserializer): Promise<void>;
    query<T>(sql: string, type?: ReceiveType<T>): Promise<any>;
    abstract execute(host: Host, connection: PostgresClientConnection, transaction?: PostgresDatabaseTransaction): Promise<T>;
    needsWritableHost(): boolean;
    handleResponse(response: Uint8Array): void;
}
export declare class InsertCommand extends Command<number> {
    platform: DefaultPlatform;
    prepared: PreparedEntity;
    items: any[];
    constructor(platform: DefaultPlatform, prepared: PreparedEntity, items: any[]);
    execute(host: Host, client: PostgresClientConnection, transaction?: PostgresDatabaseTransaction): Promise<number>;
}
export declare class FindCommand<T> extends Command<T[]> {
    sql: string;
    prepared: PreparedEntity;
    selector: SelectorState;
    private statementName;
    cache?: Buffer;
    deserialize: (message: Uint8Array, rows: any[]) => any;
    created: boolean;
    constructor(sql: string, prepared: PreparedEntity, selector: SelectorState, statementName: string);
    setParameters(params: any[]): void;
    execute(host: Host, connection: PostgresClientConnection, transaction?: PostgresDatabaseTransaction): Promise<any[]>;
}
export declare class PostgresStats {
    /**
     * How many connections have been created.
     */
    connectionsCreated: number;
    /**
     * How many connections have been reused.
     */
    connectionsReused: number;
    /**
     * How many connection requests were queued because pool was full.
     */
    connectionsQueued: number;
    bytesReceived: number;
    bytesSent: number;
}
export declare class PostgresConnectionPool {
    protected config: PostgresClientConfig;
    protected stats: PostgresStats;
    protected connectionId: number;
    /**
     * Connections, might be in any state, not necessarily connected.
     */
    connections: PostgresClientConnection[];
    protected cacheHints: {
        [key: string]: PostgresClientConnection;
    };
    protected queue: {
        resolve: (connection: PostgresClientConnection) => void;
        request: ConnectionRequest;
    }[];
    protected lastError?: Error;
    constructor(config: PostgresClientConfig, stats: PostgresStats);
    protected waitForAllConnectionsToConnect(throws?: boolean): Promise<void>;
    connect(): Promise<void>;
    close(): void;
    protected ensureHostsConnectedPromise?: Promise<void>;
    ensureHostsConnected(throws?: boolean): Promise<void>;
    protected findHostForRequest(hosts: Host[], request: ConnectionRequest): Host;
    protected createAdditionalConnectionForRequest(request: ConnectionRequest): PostgresClientConnection;
    protected newConnection(host: Host): PostgresClientConnection;
    protected release(connection: PostgresClientConnection): void;
    protected matchRequest(connection: PostgresClientConnection, request: ConnectionRequest): boolean;
    /**
     * Returns an existing or new connection, that needs to be released once done using it.
     */
    getConnection(request?: Partial<ConnectionRequest>, cacheHint?: string): Promise<PostgresClientConnection>;
}
export declare class PostgresClientConnection {
    id: number;
    readonly host: Host;
    protected config: PostgresClientConfig;
    protected onClose: (connection: PostgresClientConnection) => void;
    protected onRelease: (connection: PostgresClientConnection) => void;
    protected onSent: (bytes: number) => void;
    protected onReceived: (bytes: number) => void;
    protected messageId: number;
    status: ConnectionStatus;
    bufferSize: number;
    cache: {
        [key: string]: Command<unknown>;
    };
    released: boolean;
    connectingPromise?: Promise<void>;
    lastCommand?: {
        command: Command<unknown>;
        promise?: Promise<any>;
    };
    activeCommands: number;
    executedCommands: number;
    activeTransaction: boolean;
    reserved: boolean;
    cleanupTimeout: any;
    protected socket: Socket | TLSSocket;
    transaction?: PostgresDatabaseTransaction;
    responseParser: ResponseParser;
    error?: Error;
    bytesReceived: number;
    bytesSent: number;
    statementId: number;
    protected boundWrite: (message: Uint8Array) => void;
    constructor(id: number, host: Host, config: PostgresClientConfig, onClose: (connection: PostgresClientConnection) => void, onRelease: (connection: PostgresClientConnection) => void, onSent: (bytes: number) => void, onReceived: (bytes: number) => void);
    getCache(key: string): Command<unknown> | undefined;
    setCache(key: string, command: Command<unknown>): void;
    isConnected(): boolean;
    isConnecting(): boolean;
    close(): void;
    run(sql: string): Promise<any>;
    query(sql: string, params: any): any;
    prepare(sql: string, selector: SelectorState): PostgresClientPrepared;
    release(): void;
    /**
     * When a full message from the server was received.
     */
    protected onResponse(message: Uint8Array): void;
    /**
     * Puts a command on the queue and executes it when queue is empty.
     * A promise is return that is resolved when executed successfully, or rejected
     * when timed out, parser error, or any other error.
     */
    execute<T extends Command<unknown>>(command: T): Promise<ReturnType<T['execute']>>;
    write(message: Uint8Array): void;
    connect(): Promise<void>;
}
export declare class PostgresClient {
    pool: PostgresConnectionPool;
    stats: PostgresStats;
    config: PostgresClientConfig;
    constructor(options: PostgresClientConfig | string);
    /**
     * Returns an existing or new connection, that needs to be released once done using it.
     */
    getConnection(request?: Partial<ConnectionRequest>, transaction?: PostgresDatabaseTransaction): Promise<PostgresClientConnection>;
}
export {};
export declare type __ΩResult = any[];
export declare type __ΩConnectionRequest = any[];
export declare type __ΩTransactionTypes = any[];
