"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2021 Deepkit UG, Marc J. Schmidt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the MIT License.
 *
 * You should have received a copy of the MIT License along with this program.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostgresPlatform = exports.PostgresPlaceholderStrategy = void 0;
const sql_1 = require("@deepkit/sql");
const postgres_serializer_js_1 = require("./postgres-serializer.js");
const type_1 = require("@deepkit/type");
const postgres_schema_parser_js_1 = require("./postgres-schema-parser.js");
const sql_filter_builder_js_1 = require("./sql-filter-builder.js");
const core_1 = require("@deepkit/core");
const sqlstring_1 = __importDefault(require("sqlstring"));
function escapeLiteral(value) {
    if (value === null || value === undefined)
        return 'null';
    if (value instanceof Date)
        return sqlstring_1.default.escape(value);
    if ('number' === typeof value || 'bigint' === typeof value)
        return String(value);
    if ('string' !== typeof value)
        return escapeLiteral(String(value));
    let hasBackslash = false;
    let escaped = '\'';
    for (let i = 0; i < value.length; i++) {
        const c = value[i];
        if (c === '\'') {
            escaped += c + c;
        }
        else if (c === '\\') {
            escaped += c + c;
            hasBackslash = true;
        }
        else {
            escaped += c;
        }
    }
    escaped += '\'';
    if (hasBackslash) {
        escaped = ' E' + escaped;
    }
    return escaped;
}
class PostgresPlaceholderStrategy extends sql_1.SqlPlaceholderStrategy {
    getPlaceholder() {
        return '$' + (++this.offset);
    }
}
exports.PostgresPlaceholderStrategy = PostgresPlaceholderStrategy;
class PostgresPlatform extends sql_1.DefaultPlatform {
    constructor() {
        super();
        this.defaultSqlType = 'text';
        this.annotationId = 'postgres';
        this.defaultNowExpression = `now()`;
        this.serializer = postgres_serializer_js_1.postgresSerializer;
        this.schemaParserType = postgres_schema_parser_js_1.PostgresSchemaParser;
        this.placeholderStrategy = PostgresPlaceholderStrategy;
        this.addType(() => true, 'jsonb'); //default everything is jsonb
        this.addType(sql_1.typeResolvesToNumber, 'double precision');
        this.addType(sql_1.typeResolvesToInteger, 'integer');
        this.addType(sql_1.typeResolvesToBigInt, 'bigint');
        this.addType(sql_1.typeResolvesToBoolean, 'boolean');
        this.addType(sql_1.typeResolvesToString, 'text');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.integer, 'integer');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.int8, 'smallint');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.uint8, 'smallint');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.int16, 'smallint');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.uint16, 'smallint');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.int32, 'integer');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.uint32, 'integer');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.float32, 'real');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.float64, 'double precision');
        this.addType(type => type.kind === type_1.ReflectionKind.number && type.brand === type_1.TypeNumberBrand.float, 'double precision');
        this.addType(type_1.isUUIDType, 'uuid');
        this.addBinaryType('bytea');
        this.addType(sql_1.typeResolvesToDate, 'timestamp');
    }
    getSqlTypeCaster(type) {
        if ((0, type_1.isReferenceType)(type)) {
            type = type_1.ReflectionClass.from(type).getPrimary().type;
        }
        const dbType = this.getTypeMapping(type);
        if (!dbType)
            return super.getSqlTypeCaster(type);
        return (placeholder) => placeholder + '::' + dbType.sqlType;
    }
    getAggregateSelect(tableName, property, func) {
        if (func === 'group_concat') {
            return `array_to_string(array_agg(${tableName}.${this.quoteIdentifier(property.name)}), ',')`;
        }
        return super.getAggregateSelect(tableName, property, func);
    }
    createSqlFilterBuilder(adapter, schema, tableName) {
        return new sql_filter_builder_js_1.PostgreSQLFilterBuilder(adapter, schema, tableName, this.serializer, new this.placeholderStrategy);
    }
    getDeepColumnAccessor(table, column, path) {
        return `${table ? table + '.' : ''}${this.quoteIdentifier(column)}->${this.quoteValue(path)}`;
    }
    quoteValue(value) {
        if (!(value instanceof Date) && ((0, core_1.isObject)(value) || (0, core_1.isArray)(value)))
            return escapeLiteral(JSON.stringify(value));
        if (value instanceof Date)
            return 'TIMESTAMP ' + sqlstring_1.default.escape(value);
        return escapeLiteral(value);
    }
    getColumnType(column) {
        const ddl = [];
        if (column.isAutoIncrement) {
            ddl.push(`SERIAL`);
        }
        else {
            const foreignKey = column.table.getForeignKeyOfLocalColumn(column);
            if (foreignKey) {
                const [foreignPk] = foreignKey.foreignColumns;
                if (foreignPk.isAutoIncrement) {
                    //the foreignPK has `SERIAL` type, which is basically an INTEGER.
                    ddl.push('INTEGER');
                }
                else {
                    ddl.push((foreignPk.type || 'INTEGER') + column.getSizeDefinition());
                }
            }
            else {
                ddl.push((column.type || 'INTEGER') + column.getSizeDefinition());
            }
        }
        return ddl.filter(sql_1.isSet).join(' ');
    }
    getColumnDDL(column) {
        const ddl = [];
        ddl.push(this.getIdentifier(column));
        ddl.push(this.getColumnType(column));
        if (!column.isAutoIncrement) {
            ddl.push(column.isNotNull ? this.getNotNullString() : this.getNullString());
            ddl.push(this.getColumnDefaultValueDDL(column));
        }
        return ddl.filter(sql_1.isSet).join(' ');
    }
    getModifyColumnDDL(diff) {
        // postgres doesn't support multiple column modifications in one ALTER TABLE statement
        // see https://www.postgresql.org/docs/current/sql-altertable.html
        const lines = [];
        const identifier = this.getIdentifier(diff.to);
        if (diff.from.type !== diff.to.type || diff.from.isAutoIncrement !== diff.to.isAutoIncrement) {
            lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} TYPE ${this.getColumnType(diff.to)}`);
        }
        if (diff.from.isNotNull !== diff.to.isNotNull) {
            if (diff.to.defaultExpression !== undefined || diff.to.defaultValue !== undefined) {
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} SET ${this.getColumnDefaultValueDDL(diff.to)}`);
            }
            else {
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} DROP DEFAULT`);
            }
        }
        if (diff.from.isNotNull !== diff.to.isNotNull) {
            if (diff.to.isNotNull) {
                //NOT NULL is newly added, so we need to update all existing rows to have a value
                const defaultExpression = this.getDefaultExpression(diff.to);
                if (defaultExpression) {
                    lines.push(`UPDATE ${this.getIdentifier(diff.to.table)} SET ${identifier} = ${defaultExpression} WHERE ${identifier} IS NULL`);
                }
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} SET NOT NULL`);
            }
            else {
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} DROP NOT NULL`);
            }
        }
        return lines.join(';\n');
    }
    getDefaultExpression(column) {
        if (undefined !== column.defaultValue) {
            // if type is from JSON type, we need to cast whatever is in defaultValue to JSON
            if (column.type === 'jsonb')
                return `${this.quoteValue(JSON.stringify(column.defaultValue))}::jsonb`;
            if (column.type === 'json')
                return `${this.quoteValue(JSON.stringify(column.defaultValue))}::json`;
        }
        return super.getDefaultExpression(column);
    }
    getUniqueDDL(unique) {
        return `CONSTRAINT ${this.getIdentifier(unique)} UNIQUE (${this.getColumnListDDL(unique.columns)})`;
    }
    getDropIndexDDL(index) {
        if (index.isUnique) {
            return `ALTER TABLE ${this.getIdentifier(index.table)} DROP CONSTRAINT ${this.getIdentifier(index)}`;
        }
        return super.getDropIndexDDL(index);
    }
    supportsInlineForeignKey() {
        return false;
    }
    supportsAggregatedAlterTable() {
        return false;
    }
    supportsSelectFor() {
        return true;
    }
    getAutoIncrement() {
        return '';
    }
    getDropTableDDL(table) {
        return `DROP TABLE IF EXISTS ${this.getIdentifier(table)} CASCADE`;
    }
    getUseSchemaDDL(table) {
        if (!table.schemaName)
            return '';
        return `SET search_path TO ${this.quoteIdentifier(table.schemaName)}`;
    }
    getResetSchemaDDL(table) {
        if (!table.schemaName)
            return '';
        return `SET search_path TO public`;
    }
}
exports.PostgresPlatform = PostgresPlatform;
//# sourceMappingURL=postgres-platform.js.map